"""  The :obj:`stream` submodule: representing the sound signal

Containing the ``Stream`` class to house the ``Sonificiation`` audio
signal for each channel in the ``Channels`` object. This can be
split into uniform segments or `buffers` via the ``Buffers`` object,
for processing.

Todo:
  * implement filter Q-parameter mapping
"""
import numpy as np
import wavio
import matplotlib.pyplot as plt
from scipy.signal.windows import hann

class Stream:
    """
    Stream object representing audio samples.

    Houses audio samples and associates metadata representing the
    actual audio signal produced by the `Generator` class and output
    via the `audio_channels` class.

    Attributes:
      samprate (:obj:`int`): Samples per second of audio stream (Hz)
      length (:obj:`float`): Duration of the stream in seconds
      values (:obj:`ndarray`): Values of individual samples
      samples (:obj:`ndarray`): Indices of each sample
      samptype (:obj:`ndarray`): Time in seconds each sample occurs
      buffers (:obj:`Buffers`): Buffered stream if generated
    """
    def __init__(self, length, samprate=44100, ltype='seconds'):
        """ 
        Args:
          length (numerical): Number representing the length of the
            stream either as an integer number of samples, or a value
            of seconds
          samprate (optional :obj:`int`): Samples per second of audio
            stream (Hz)
          ltype (optional :obj:`str`): quantity represented by
            `length`, either duration in 'seconds' or precise number 
            of 'samples'
        """
        # variables we want to keep constant
        self.samprate = samprate
        self._nyqfrq = 0.5*self.samprate
        
        if ltype == 'seconds':
            self.length = length
            self._nsamp_stream = int(samprate * length)
        elif ltype == 'samples':
            self._nsamp_stream = length
            self.length = length / samprate
            
            
        # sample values initialised to 0 (silence)
        self.values =  np.zeros(self._nsamp_stream)

        # private stream for keeping track of buffered stream
        self._bvalues =  np.zeros(self._nsamp_stream)
        
        # sample numbers for indexing
        self.samples = np.arange(self._nsamp_stream, dtype=int)

        # time at which each sample occurs
        self.samptime = self.samples / self.samprate
        
    def bufferize(self, bufflength=0.1):
        """Wrapper to initialise Buffers subclass

        Args:
          bufflength (optional, :obj:`float`): duration in seconds of
            each buffer to be generated
        """
        self.buffers = Buffers(self, bufflength)

    def consolidate_buffers(self):
        """
        Wrapper to reassign stream values to consolidated buffers

        See :func:`~stream.Buffers.buffers.to_stream`
        """
        self.values = self.buffers.to_stream()
        
    def filt_sweep(self, ffunc, fmap, qmap=lambda x:x*0 + 0.1,
                   flo=20, fhi=2.205e4, qlo=0.5, qhi=10):
        """
        Apply time varying filter to buffered stream

        Args:
          ffunc (function): function that applies filter 
          fmap (function): mapping function representing filter cutoff
            sweep
          qmap (optional, function): mapping function for a filters
            Q parameter
          flo (optional, :obj:`float`): lowest frequency of sweep in Hz,
            default 20
          fhi (optional, :obj:`float`): lowest frequency of sweep in Hz,
            default 22.05 kHz
          qlo (optional, :obj:`float`): lowest 'Q' value of sweep,
            default 0.5
          qhi (optional, :obj:`float`): lowest frequency of sweep,
            default 10
          
        """
        if not hasattr(self, "buffers"):
            Exception("needs bufferized stream, please run 'bufferize' method first.")
        buffers = self.buffers

        # array to regularly sample maps at each buffer
        x = np.linspace(0, 1, buffers._nbuffs_tot)

        lfhi = np.log10(fhi)
        lflo = np.log10(flo)
        
        # obtain buffer values from maps, with cutoff sweep in units
        # of the nyquist frequency
        svals = pow(10., fmap(x)*(lfhi-lflo)+lflo)/self._nyqfrq
        qvals = (qmap(x)*(qhi-qlo)+qlo)

        # loop over buffers, applying appropriate filtering to each 
        for i in range(buffers._nbuffs):
            i2 = 2*i
            buffers.buffs_tile[i] = ffunc(buffers.buffs_tile[i], svals[i2], qvals[i2])
        for i in range(buffers._nbuffs-1):
            i2 = 2*i+1
            buffers.buffs_olap[i] = ffunc(buffers.buffs_olap[i], svals[i2], qvals[i2])

        # finally, consolidate buffers to apply effect to stream
        self.consolidate_buffers()

    def get_sampfracs(self):
        """ Get fractional position of the sample in total stream duration
        """
        self.sampfracs = np.linspace(0, 1, self.values.size)
        
    def save_wav(self, filename):
        """Save audio stream to wav file, specified by filename

        Args:
          filename (:obj:`str`): name of output WAV file 
        """
        wavio.write(filename, self.values, self.samprate, sampwidth=3) 
        
    def reset(self):
        """Zero audio stream and buffers if present."""
        self.values *= 0.
        if hasattr(self, "buffers"):
            self.buffers.buffs_tile *= 0.
            self.buffers.buffs_olap *= 0.        
        
class Buffers:
    """Audio buffers split into uniform discrete chunks or 'buffers'.

    Audio ~:class:`stream.Stream` as a discrete sequence of individual
    'buffers' of fixed duration (number of samples). This allows time
    varying operations in frequency space, such as signal filtering.
    Buffers are tiled in a 'brickwork' fashion so they always overlap
    with another buffer.

    Attributes:
      fade (:obj:`ndarray`): Window function for recombining
        overlapping buffers
      nsamp_padstream (:obj:`int`): Number of samples needed to split
        the stream into discrete buffers of chosen length
      nsamp_pad (:obj:`int`): Number of additional samples needed to
        add to the original `Stream` size in this case
      buffs_tile (:obj:`ndarray`): 2d array of buffers completely
        enclosing the stream (number of buffers x samples per buffer)
      buffs_olap (:obj:`ndarray`) 2d array of overlap buffers,
        allowing for cross fading 
    """
    def __init__(self, stream, bufflength=0.1):
        """
        Args:
          stream (~:class:`stream.Stream`): Stream object to be
            represented using the buffers
          bufflength (optional, :obj:`float`): duration in seconds of
            each buffer to be generated
        """
        nbuff = stream.samprate*bufflength
        if nbuff < 20:
            Exception(f"Error: buffer length {nbuff} samples below "
                      "lower limit of 20, with specified bufflength "
                      "{bufflength} seconds and sample rate {self.samprate} Hz")

        # force buffer length to an even number of samples
        self._nsamp_halfbuff = int(stream.samprate*bufflength) // 2
        self._nsamp_buff = 2 * self._nsamp_halfbuff

        # minimum number of tiled buffers to completely enclose stream 
        self._nbuffs = 1+(stream._nsamp_stream//self._nsamp_buff)

        # total number buffers including overlaps
        self._nbuffs_tot = 2*self._nbuffs-1

        # tent function for linearly x-fading buffers on recombination
        # self.fade = 1.-abs(np.linspace(1,-1, self._nsamp_buff))
        self.fade = hann(self._nsamp_buff)

        # pad the stream up to an exact multiple of buffer sample length
        self.nsamp_padstream = self._nbuffs * self._nsamp_buff
        self.nsamp_pad = self.nsamp_padstream-stream._nsamp_stream
        self._olap_pad = self.nsamp_pad-self._nsamp_halfbuff
        self._olap_lim = min(stream._nsamp_stream, stream._nsamp_stream+self._olap_pad)
        
        # construct tile and overlap buffer arrays
        self.buffs_tile = np.pad(stream.values, (0,self.nsamp_pad)
                                 ).reshape((self._nbuffs, self._nsamp_buff))
        self.buffs_olap = np.pad(stream.values[self._nsamp_halfbuff:self._olap_lim],
                                 (0,max(0, self._olap_pad))
                                 ).reshape((self._nbuffs-1), self._nsamp_buff)

    def to_stream(self):
        """Reconstruct stream by cross-fading buffers

        Takes the `self.buffs_tile` and `self.buffs_olap` arrays and using
        the `self.fade` window function, add overlapping sample values
        together to yield a 1d array of samples.

        Returns:
          out (:obj:`ndarray`): 1d array of sample values representing the
            new audio signal for the parent `Stream`. 
        """
        # apply fades to buffers, first special edge cases...
        self.buffs_tile[0,self._nsamp_halfbuff:] *= self.fade[self._nsamp_halfbuff:]
        self.buffs_tile[-1,:self._nsamp_halfbuff] *= self.fade[:self._nsamp_halfbuff]

        # ...then to remaining buffers
        self.buffs_tile[1:-1,:] *= self.fade.T
        self.buffs_olap *= self.fade.T

        # reconstruct stream
        padded_stream = np.zeros(self.nsamp_padstream)
        padded_stream += self.buffs_tile.flatten()
        flat_olaps    = self.buffs_olap.flatten()
        padded_stream[self._nsamp_halfbuff:-self._nsamp_halfbuff] += flat_olaps

        # remove padding on returning reconstructed stream
        return padded_stream[:-self.nsamp_pad]



""" The :obj:`generator` submodule: creating sounds for the sonification.

This submodule handles the actual generation of sound for the
sonification, after parameterisation by the :obj:`Sources` and musical
choices dictated by the :obj:`Score`.

Todo:
    * Consolidate more common code into the :obj:`Generator` parent
      class.
    * Support more Envelope and LFO types in the :obj:`play` methods
      (want pitch, volume and filter options for each)
    * Check buffer length consistency for spectralizer - do we hit
      grid points?
    * Throw appropriate errors when rendering with unreasonable length
      and freq combinations
"""

from . import stream
from . import notes
from . import presets
from . import utilities as utils
from . import filters
import numpy as np
import scipy
# can we use FFTW backend in scipy?
try:
    import pyfftw
    scipy.fft.set_backend(pyfftw.interfaces.scipy_fft)
except (OSError, ModuleNotFoundError):
    pass
from scipy.fft import fft, ifft, fftfreq
import glob
import copy
import scipy
import json
from scipy.io import wavfile
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt
import warnings
import logging
from sf2utils.sf2parse import Sf2File
from pathlib import Path
import os

# ignore wavfile read warning that complains due to WAV file metadata
warnings.filterwarnings("ignore", message="Chunk \(non-data\) not understood, skipping it\.")

# TO DO:
# - Ultimately have Synth and Sampler classes that own their own stream (stream.py) object
#   allowing ADSR volume and filter enveloping, LFO implementation etc.
# - Functions here will generally be called from a "Score" class that is provided with the
#   musical choices and uses these to generate sound, but can be interfaced with directly.

def forward_loopsamp(s, start, end):
    """Produce array of sample indices for looping a sample forward.

    Sample indices between values `start` and `end` that will loop the sample
    such that it loops "forward", i.e. start, start+1, ..., end-1, end, start,
    ... etc.

    Args:
      s (:obj:`ndarray`): array of input sample indices
      start (:obj:`int`): Index of sample after which looping should commence
      end (:obj:`int`): Index of sample after which audio loops

    Returns:
      out (:obj:`ndarray`): array of output sample indices
    """
    delsamp = end-start
    return np.piecewise(s, [s < start, s >= start],
                        [lambda x: x, lambda x: (x-start)%(delsamp) + start])
def forward_back_loopsamp(s, start, end):
    """Produce array of sample indices for looping a sample forward-back.

    Sample indices between values `start` and `end` that will loop the sample
    such that it loops "forward-back", i.e. `start, start+1, ..., end-1, end,
    end-1, ..., start+1, start, start+1, ...` etc.
    ... etc.

    Args:
      s (:obj:`ndarray`): array of input sample indices
      start (:obj:`int`): Index of sample after which looping should commence
      end (:obj:`int`): Index of sample after which audio loops

    Returns:
      out (:obj:`ndarray`): array of output sample indices
    """
    delsamp = end-start
    return np.piecewise(s, [s < start, s >= start],
                        [lambda x: x, lambda x: end - abs((x-start)%(2*(delsamp)) - (delsamp))])

class Generator:
    """Generic generator Class, defining common code for child classes

    Generators have common initialisation and methods that are
    defined by this parent class.

    Attributes:
      samprate (:obj:`int`): Samples per second of audio stream (Hz)
      audbuff (:obj:`int`): Samples per audio buffer
      preset (:obj:`dict`): Dictionary of parameters defining the
        generator.

    """
    def __init__(self, params={}, samprate=48000):
        """
        Args:
    	params (`optional`, :obj:`dict`): any generator parameters
    	  that differ from the generator :obj:`preset`, where keys and
    	  values are parameter names and values respectively. 
    	samprate (`optional`, :obj:`int`): the sample rate of
  	  the generated audio in samples per second (Hz)
        """
        self.samprate = samprate

        # samples per buffer (use 30Hz as minimum)
        self.audbuff = self.samprate / 30.
        
        # modify or load preset if specified
        if params:
            self.preset = self.modify_preset(params)

    def load_preset(self, preset='default'):
        """Load parameters from a preset YAML file.

        Wrapper method for the :obj:`presets.synth.load_preset` or
        :obj:`presets.sampler.load_preset` functions. Always load the
        :obj:`default` preset first to ensure all parameters are
        defined, and then if necessary reload parameters defined by
        :obj:`preset`

        Args:
          preset (:obj:`str`): name of the preset. Built-in presets
            can be named directly and looks to import the preset from
            the :obj:`<MODULE_PATH>/presets/<GENERATOR>/` directory as
            :obj:`<preset>.yml`, where :obj:`<GENERATOR>` is either
            `synth` or `sampler`, :obj:`<MODULE_PATH>` is the path to
            the strauss module (i.e. :obj:`strauss.__file__`). Custom
            presets can also  be loaded from :obj:`<preset>.yml` if
            :obj:`preset` represents a path containing file
            separators.
        """
        if not hasattr(self, preset):
            self.preset = getattr(presets, self.gtype).load_preset('default')
            self.preset['ranges'] = getattr(presets, self.gtype).load_ranges() 
        if preset != 'default':
            preset = getattr(presets, self.gtype).load_preset(preset)
            self.modify_preset(preset)

    def modify_preset(self, parameters, cleargroup=[]):
        """Modify parameters within current preset

        method allows user to tweak generator parameters directly,
        using a dictionary of parameters and their values. subgroups
        within the preset are represented as nested dictionaries. 

        Args:
          parameters (:obj:`dict`): keys and items are the preset
            parameter names and new values. Nested dictionaries are
            used to redefine grouped parameters,
            e.g. :obj:`{'volume_envelope':{'A':0.5}}`
          cleargroup (optional :obj:`list(str)`): if required, list of
            the group names to be completely reset (e.g. if defining
            new :obj:`oscillators` set for synth). 
        """
        utils.nested_dict_reassign(parameters, self.preset)
        for grp in cleargroup:
            if grp in parameters:
                for k in list(self.preset[grp].keys()):
                    if k not in parameters[grp]:
                        del self.preset[grp][k]
            
    def preset_details(self, term="*"):
        """Print the names and descriptions of presets

        Wrapper for preset_details function. lists the name and description
        of built-in presets with names matching the search term.

        Args:
          term (optional, :obj:`str`): name or glob term for built-in
          preset(s). Default '*' prints all.
        """
        getattr(presets, self.gtype).preset_details(name=term)

    def envelope(self, samp, params, etype='volume'):
        """Envelope function for modulating a single note

        The envelope function takes the pre-defined envelope
        parameters for the specified envelope type and returns the
        envelope value at each sample. envelopes are defined by
        attack, decay, sustain and release (:obj:`'A','D','S' & 'R`)
        values, as well as segment curvatures (:obj:`'Ac','Dc', &
        'Rc`) and a normalisation :obj:`'level'`. See `this article <https://learnmusicproduction.in/blogs/music-production-and-audio-engineering/adsr-fundamentals-in-music-everything-you-need-to-know>`_ for a more detailed explanation of ADSR envelopes.

        Args:
          samp (:obj:`array-like`): Audio sample index
          params (:obj:`dict`): Keys and values of generator
            parameters
          etype (optional , :obj:`str`): type of envelope, indicating
            which :obj:`params` group to read (i.e. if
            :obj:`etype='volume'`, read `from :obj:`volume_envelope`)          
        """
        # TO DO: is it worth it to pre-set this in part if parameters don't change?
        nlen=params['note_length']
        edict=params[f'{etype}_envelope']
        
        # read envelope params from dictionary
        a = edict['A']
        d = edict['D']
        s = edict['S']
        r = edict['R']
        a_k = edict['Ac']
        d_k = edict['Dc']
        r_k = edict['Rc']
        lvl = edict['level']
        
        # effective input sample times, clipped to ensure always defined
        sampt = samp/self.samprate
        
        # handy time values
        t1 = a 
        t2 = a+d
        t3 = nlen+r

        # determine segments and envelope value when note turns off
        a_seg = lambda t: 1-self.env_segment_curve(t, a, 1, -a_k)
        d_seg = lambda t: s+self.env_segment_curve(t-t1, d, 1-s, d_k)
        s_seg = lambda t: s
        o_seg = lambda t: 0.

        if nlen < t1:
            env_off = a_seg(nlen)
        elif nlen < t2:
            env_off = d_seg(nlen)
        else:
            env_off = s

        r_seg = lambda t: self.env_segment_curve(t-nlen, r, env_off, r_k)

        # conditionals to determine which segment a sample is in
        a_cond = sampt < t1
        d_cond = np.logical_and(sampt<min(t2,nlen), sampt>=t1)
        s_cond = np.logical_and(sampt<nlen, sampt>=min(t2,nlen))
        r_cond = sampt >= nlen
        o_cond = sampt >= t3

        # compute envelope for each sample 
        env =  np.piecewise(sampt,
                            [a_cond, d_cond, s_cond, r_cond, o_cond],
                            [a_seg, d_seg, s_seg, r_seg, o_seg])
        return lvl*env

    def env_segment_curve(self, t, t1, y0, k):
        """Formula for segments of the envelope function
        
        Function to evaluate the segments of the envelope, allowing
        for curvature, i.e. concave & convex envelope segments.

        Args:
          t (:obj:`float`): time of each sample along segment
          t1 (:obj:`float`): time of segment endpoint
          y0 (:obj:`float`): starting value of segment
          k (:obj:`float`): curvature value of segment, from (-1,1),
            with positive values indicating concave and negative
            convex curvature.
        """
        return y0/(1 + (1-k)*t / ((k+1)*(t1-t)))

    
    # ||||||||||||||||||||||||||||||||||||||||||||||||||
    # OSC types 
    # ||||||||||||||||||||||||||||||||||||||||||||||||||
    def sine(self, s,f,p):
        """Sine-wave oscillator

        Args:
          s (:obj:`array`-like): sample index
          f (:obj:`float`): samples per cycle
          p (:obj:`float` or :obj:`str`): if numerical, phase in units
            of cycles, :obj:`'random'` indicates randomised.
        Returns:
          v (:obj:`array`-like): values for each sample
        """
        return np.sin(2*np.pi*(s*f+p))
    
    def saw(self,s,f,p):
        """Sawtooth-wave oscillator

        Args:
          s (:obj:`array`-like): sample index
          f (:obj:`float`): samples per cycle
          p (:obj:`float` or :obj:`str`): if numerical, phase in units
            of cycles, :obj:`'random'` indicates randomised.
        Returns:
          v (:obj:`array`-like): values for each sample
        """
        return (2*(s*f+p) +1) % 2 - 1
    
    def square(self,s,f,p):
        """Square-wave oscillator

        Args:
          s (:obj:`array`-like): sample index
          f (:obj:`float`): samples per cycle
          p (:obj:`float` or :obj:`str`): if numerical, phase in units
            of cycles, :obj:`'random'` indicates randomised.
        Returns:
          v (:obj:`array`-like): values for each sample
        """
        return np.sign(self.saw(s,f,p))
    
    def tri(self,s,f,p):
        """Triangle-wave oscillator

        Args:
          s (:obj:`array`-like): sample index
          f (:obj:`float`): samples per cycle
          p (:obj:`float` or :obj:`str`): if numerical, phase in units
            of cycles :obj:`'random'` indicates randomised.
        Returns:
          v (:obj:`array`-like): values for each sample
        """
        return 1 - abs((4*(s*f+p) +1) % 4 - 2)
    def noise(self,s,f,p):
        """White noise oscillator

        Note:
          :obj:`f` and :obj:`p` have no efffect for this oscillator,
          generating a random value for each sample.
        Args:
          s (:obj:`array`-like): sample index
          f (:obj:`float`): unused
          p (:obj:`float` or :obj:`str`): unused
        Returns:
          v (:obj:`array`-like): values for each sample
        """
        return np.random.random(np.array(s).size)*2-1

    def lfo(self, samp, sampfrac, params, ltype='pitch'):
        """Low-Frequency oscillator (LFO)

        This function takes the pre-defined LFO parameters (if
        switched on) for the specified LFO type and returns the
        LFO value at each sample. LFOs are defined by the same values
        as :meth:`strauss.generator.envelope`, with an additional
        :obj:`use` switch, a waveform (:obj:`wave`,
        e.g. :obj:`sine`), an amplitude (:obj:`amount`), a frequncy in
        Hz (:obj:`freq`) a frequency shift in octaves
        (:obj:`freq_shift`) and a :obj:`phase`, either numerical in
        cycles or :obj:`'random'` to indicate randomised.

        Note:
          To modulate the frequency of an oscillator, use the
          :obj:`freq_shift` parameter, rather than :obj:`freq`

        Args:
          samp (:obj:`array-like`): Audio sample index
          sampfrac (:obj:`array-like`): Audio sample as fraction of
            total number of samples
          params (:obj:`dict`): Keys and values of generator
            parameters
          ltype (optional , :obj:`str`): type of LFO, indicating which
            :obj:`params` group to read (i.e. if :obj:`ltype='volume',
            read `from :obj:`pitch_envelope`)

        Returns:
          v (:obj:`array-like`): amplitude of LFO at each input sample
        """
        env_dict = {}
        lfo_key = f'{ltype}_lfo'
        lfo_params = params[lfo_key]
        
        env_dict['note_length'] = params['note_length']
        env_dict['lfo_envelope'] = lfo_params

        freq = lfo_params['freq']/self.samprate
        effsamp = samp.astype(float)
        
        if callable(lfo_params['freq_shift']):
            findex = lfo_params['freq_shift'](sampfrac)
            effsamp = np.cumsum(pow(2, findex))
        elif lfo_params['freq_shift'] != 0:
            effsamp *=  pow(2,lfo_params['freq_shift'])
            
        if callable(lfo_params['amount']):
            amnt  = lfo_params['amount'](sampfrac)
        else:
            amnt = lfo_params['amount']

        if lfo_params['phase'] == 'random':
            phase = np.random.random()
        else:
            phase =lfo_params['phase']

        osc = getattr(self,lfo_params['wave'])(effsamp, freq, phase)
        env = self.envelope(samp, env_dict, 'lfo')
        return amnt * env * osc
        
class Synthesizer(Generator):
    """Synthesizer generator class

    This generator class synthesises sound using mathmatically
    generated waveforms or 'oscillators', from a combination of
    oscillator methods defined in the parent class. The relative
    frequency, phase and amplitude of these oscillators are defined in
    the preset, and linearly combined to produce the sound. defines
    attribute :obj:`self.gtype = 'synth'`.

    Attributes:
      gtype (:obj:`str`): Generator type 

    Todo:
    	* Add other synthesiser types, aside from additive (e.g. FM,
    	  vector, wavetable)? 
    """
    def __init__(self, params=None, samprate=48000):
        """
        Args:
    	  params (`optional`, :obj:`dict`): any generator parameters
    	    that differ from the generator :obj:`preset`, where keys
            and values are parameters names and values respectively. 
    	  samprate (`optional`, :obj:`int`): the sample rate of
  	    the generated audio in samples per second (Hz)
        """
        # default synth preset
        self.gtype = 'synth'
        self.preset = getattr(presets, self.gtype).load_preset()
        self.preset['ranges'] = getattr(presets, self.gtype).load_ranges() 
        
        # universal initialisation for generator objects:
        super().__init__(params, samprate)

        # set up the oscillator banks
        self.setup_oscillators()

    def setup_oscillators(self):
        """Setup and consolidate oscs into a two-variable function.

        Reads the parameterisation of each oscillator from the preset,
        specifying their waveform (:obj:`wave`), relative amplitude
        (:obj:`level`), detuning in cents (:obj:`det`) and
        :obj:`phase`, either a number in units of cycles, or a string
        specifying randomisation (:obj:`'random'`). Sets the
        :obj:`self.generate` method, using the
        :obj:`self.combine_oscs`.

        Note:
          This is deprecated and will likely be removed from future
          versions
        """
        # oscdict = self.preset['oscillators']
        # self.osclist = []
        # for osc in oscdict.keys():
        #     lvl = oscdict[osc]['level']
        #     det = oscdict[osc]['detune']
        #     phase = oscdict[osc]['phase']
        #     form = oscdict[osc]['form']
        #     snorm = self.samprate
        #     fnorm = (1 + det/100.)
        #     if phase == 'random':
        #         oscf = lambda samp, f: lvl * getattr(self,form)(samp/snorm, f*fnorm, np.random.random())
        #     else:
        #         oscf = lambda samp, f: lvl * getattr(self,form)(samp/snorm, f*fnorm, phase)
        #     self.osclist.append(oscf)
        #     flg += 1
        self.generate = self.combine_oscs

    def modify_preset(self, parameters, clear_oscs=True):
        """Synthesizer-specific wrapper for the modify_preset method.

        This gives control over whether or not to clear the arbitrary
        number of oscillators for synthesizer.
        
        Args:
          parameters (:obj:`dict`): keys and items are the preset
            parameter names and new values. Nested dictionaries are
            used to redefine grouped parameters,
            e.g. :obj:`{'volume_envelope':{'A':0.5}}`
          clear_oscs (optional, :obj:`bool`): if True, clear all
            oscillators from the existing preset. Turn off if just
            wishing to tweak non-oscillator parameters.
        """
        if clear_oscs:
            super().modify_preset(parameters, ['oscillators'])
        else:
            super().modify_preset(parameters)
        self.setup_oscillators()
            
    def combine_oscs(self, s, f):
        """ Evaluate and linearly combine oscillators.

        Args:
          s (:obj:`array`-like): Sample index
          f (:obj:`float` or :obj:`str`): If numerical, frequency in
            cycles per second, if string, note name in scientific pitch
            notation (e.g. :obj:`'A4'`)
        Returns:
          tot (:obj:`array`-like): values for each sample
        """
        tot = 0.
        oscdict = self.preset['oscillators']
        if isinstance(f, str):
            # we want a numerical frequency to generate tone
            f = notes.parse_note(f)
        for osc in oscdict:
            lvl = oscdict[osc]['level']
            det = oscdict[osc]['detune']
            phase = oscdict[osc]['phase']
            form = oscdict[osc]['form']
            snorm = self.samprate
            fnorm = (1 + det/100.)
            if phase == 'random':
                tot += lvl * getattr(self,form)(s/snorm, f*fnorm, np.random.random())
            else:
                tot += lvl * getattr(self,form)(s/snorm, f*fnorm, phase)
            # self.osclist.append(oscf)
            # flg += 1
        return tot

    def play(self, mapping):
        """ Play the sound for a given source.

        Play a given source and return the sample values for
        combination into the overall sonification.

        Note:
          :obj:`mapping` is a linear dictionary (not nested, as for 
          :meth:`strauss.generator.modify_preset`) where group members
          are indicated using :obj:`'/'` notation
          (e.g. :obj:`{'volume_envelope/A': 0.5, ...`).

        Args:
          mapping (:obj:`dict`): keys and items are generator
            parameter names and their values. This combines all the
            preset mapped parameters, overwritten by any
            :obj:`Source`-mapped parameters (represented as values or
            interpolation functions for static and evolving
            parameters, respectively). This is a linear dictionary
            (not nested, see :meth:`strauss.generator.modify_preset`)
            where group members are indicated using :obj:`'/'`
            notation (e.g. :obj:`{'volume_envelope/A': 0.5, ...`).

        """
        samprate = self.samprate
        audbuff = self.audbuff

        params = copy.deepcopy(self.preset)
        utils.linear_to_nested_dict_reassign(mapping, params)

        nlength = (params['note_length']+params['volume_envelope']['R'])*samprate
        
        # generator stream (attribute of stream?)
        sstream = stream.Stream(nlength/samprate, samprate)
        samples = sstream.samples
        sstream.get_sampfracs()

        pindex  = np.zeros(samples.size)
        if callable(params['pitch_shift']):
            pindex += params['pitch_shift'](sstream.sampfracs)/12.
        elif params['pitch_shift'] != 0:
            pindex += params['pitch_shift']/12.
        if params['pitch_lfo']['use']:
            pindex += self.lfo(samples, sstream.sampfracs, params, 'pitch')/12.
        if np.any(pindex):
            samples = np.cumsum(pow(2., pindex))
            # if callable(params['pitch_shift']):
            #     samples = np.cumsum(pow(2., pindex))
            # else:
            #     samples = samples * pow(2., pindex)
        
        # generate stream values
        values = self.generate(samples, params['note'])

        # get volume envelope
        env = self.envelope(sstream.samples, params)
        if params['volume_lfo']['use']:
            env *= np.clip(1.-self.lfo(sstream.samples, sstream.sampfracs,
                                       params, 'volume')*0.5, 0, 1)
        
        # apply volume normalisation or modulation (TO DO: envelope, pre or post filter?)
        sstream.values = values * utils.const_or_evo(params['volume'], sstream.sampfracs) * env

        # filter stream
        if params['filter'] == "on":
            if hasattr(params['cutoff'], "__iter__"):
                # if static cutoff, use minimum buffer count
                sstream.bufferize(sstream.length/4)
            else:
                # 30 ms buffer (hardcoded for now)
                sstream.bufferize(0.03)
            sstream.filt_sweep(getattr(filters, params['filter_type']),
                               utils.const_or_evo_func(params['cutoff']))
        return sstream    
            
class Sampler(Generator):
    """Sampler generator class

    This generator class generates sound using pre-loaded audio
    samples, representing different notes. Presets define parameters
    controlling these defines
    attribute :obj:`self.gtype = 'sampler'`.

    Attributes:
      gtype (:obj:`str`): Generator type 
    
    Todo:
    	* Add zone mapping for samples (e.g. allow a sample to define
          a range of notes played at different speeds).
        * Support non-scientifically named notes? (e.g
    	  :obj:`'cymbal'`, :obj:`'snare'`). 
        * Have sample loading defined via the preset rather than the
          :obj:`sampfiles` variable?
    """

    def __init__(self, sampfiles, params=None, samprate=48000, sf_preset=None):
        """
        Args:
          sampfiles (`required`, :obj:`str`): string pointing to samples
            to load. This can either point to a directory containing
            samples, where `"path/to/samples"` contains files named
            as `samples_A#4.wav` (ie. `<lowest_directory>_<note>.wav`),
            or a *Soundfont* file, with extension `.sf2`.
    	  params (`optional`, :obj:`dict`): any generator parameters
    	    that differ from the generator :obj:`preset`, where keys and
    	    values are parameters names and values respectively. 
    	  samprate (`optional`, :obj:`int`): the sample rate of
  	    the generated audio in samples per second (Hz)
          sf_preset (`optional`, :obj:`int`) if using a *Soundfont*
            (`.sf2`) file, this is the number of the preset to use.
            All `.sf2` files should contain at least one preset. When
            given default `None` value, will print available presets
            and select the first preset. Note presets are 1-indexed.
        
        Note:
          It is necessary to assign a note for each sample in order to
          choose different sample based on the ``pitch`` parameter. This
          is also the case for non-pitched sounds, following a similar
          approach to a [keyboard sampler]
          (https://support.apple.com/en-lk/guide/logicpro/lgcp4eecaaff/mac)
          where each key can triggers a different chosen sample. If
          `drumset_C1.wav` is a kick drum and `drumset_D1.wav` is a snare
          drum for :obj:`Score` with `chord_sequence=[["C1", "D1"]]`, events
          mapped to a higher (lower) `pitch` will sound as snare (kick) drums.
          
        """
        # default sampler preset
        self.gtype = 'sampler'
        self.preset = getattr(presets, self.gtype).load_preset()
        self.preset['ranges'] = getattr(presets, self.gtype).load_ranges() 
        
        # universal initialisation for generator objects:
        super().__init__(params, samprate)
        
        if isinstance(sampfiles, dict):
            # catch case sample dictionary provided directly
            self.sampdict = sampfiles
        else:
            # re-cast sampfiles as a string
            sampfiles = str(sampfiles)
            if sampfiles[-4:] == '.sf2':
                # if a soundfont (.sf2) file, use read routines
                with open(sampfiles, 'rb') as sf2_file:
                    self.sf2 = Sf2File(sf2_file)
                    # number of presets (excluding EOS entry)
                    npres = len(self.sf2.raw.pdta['Phdr'][:-1])
                    if npres == 1:
                        # if there's only one preset, choose it
                        sf_preset = 1
                    if not (isinstance(sf_preset, int) and (sf_preset >= 1) and (sf_preset <= npres)):
                        # if there's more than one, and no valid number specified, ask for one.
                        print("valid 'sf_preset' not provided for soundfont file, available presets are:")
                        print('\n'+''.join(['-']*40))
                        choose_name = ''
                        for i in range(npres):
                            hdr = self.sf2.raw.pdta['Phdr'][i]
                            name = json.dumps(hdr.name.decode('utf-8')).replace(r'\u0000', '')
                            print(f"{i+1}. {name}")
                            if not choose_name:
                                choose_name = name
                        print(''.join(['-']*40)+'\n')
                        # TODO: zero index, but would 1 index be more user friendly?
                        print(f"By default choosing preset 1 ({choose_name}).\n\n"
                              "Re-run 'Sampler' with the 'sf_preset' keyword argument to select a specific\n"
                              f"preset, ie. 'Sampler(\"{sampfiles}\",sf_preset=N)',\n"
                              f"where N is an integer from 1-{i+1}.\n")
                        sf_preset = 1
                    # TODO: isolate the warning suppression better?
                    logger = logging.getLogger()
                    pres = self.sf2.build_presets()
                    logger.disabled = True
                    sf_data = self.get_sfpreset_samples(pres[sf_preset-1])
                    self.sampdict = self.reconstruct_samples(sf_data)
                    logger.disabled = False
                    
            else:
                wavs = sorted(Path(sampfiles).glob("*.[wW][aA][vV]"))
                self.sampdict = {}
                for w in wavs:
                    filename = Path(w).name
                    note = filename.split('_')[-1].split('.')[0]
                    self.sampdict[note] = str(w)
        self.load_samples()

    def get_sfpreset_samples(self, sfpreset):
        """Reading samples from a soundfont file along with metadata.

        Read in the audio samples from a ``.sf2`` file to populate
        available notes, mapping the MIDI key values to musical notes,
        scaling and tuning samples as appropriate.

        Args:
          sf_preset (`optional`, :obj:`int`) The number of the *Soundfont*
            preset to use. All `.sf2` files should contain at least one
            preset. When given default `None` value, will print available
            presets and select the first preset. Note presets are
            1-indexed.

        Returns:
          sfpre_dict (:obj:`dict`): dictionary of data required to load
          soundfont samples in to the `Sampler`, including raw `samples`,
          `sample_rate`, `original_pitch` of the samples, the `min_note`
          and `max_note` in midi values to use the sample, and the
          `sample_map`, assigning each sample to a note.
        """
        minmidi = np.inf
        maxmidi = -np.inf
        stdvel = 100
        sampdat = {}
        sratedat = {}
        krangedat = {}
        mapsamps = {}
        opitchdat = {}

        # iterate through preset 'bags' containing 
        # sample sets associated to each note
        for bag in sfpreset.bags:
            isvelstd = True
            inst = bag.instrument
            vr = bag.velocity_range
            if vr:
                isvelstd = (vr[0] <= stdvel) and (vr[1] >= stdvel)
            # we support a fixed velocity, choose value stdvel
            # as standard, so only want bags of samples 
            # associated with that range
            if not isvelstd:
                continue
            if inst:
                # if bag is not empty, iterate through samples
                for sbag in inst.bags:
                # for i in range(len(inst.samples)):
                    samp = sbag.sample
                    if not samp:
                        # if sbag not associated with a sample, skip
                        continue
                    # don't support stereo samples, due to spatialisation
                    # in strauss. Only read in mono or left channel samples.  
                    if samp.is_left or samp.is_mono:
                        tune = 0
                        ftun = 0
                        if sbag.tuning:
                            tune += sbag.tuning
                        if sbag.fine_tuning:
                            tune += sbag.fine_tuning/100
                        sample = np.frombuffer(samp.raw_sample_data, dtype='int16')
                        note = samp.original_pitch
                        if sbag.base_note:
                            note = sbag.base_note
                        name = samp.name
                        keys = np.array(sbag.key_range)
                        minmidi = min(minmidi, keys[0])
                        maxmidi = max(maxmidi, keys[1])
                        for i in range(keys[0], keys[1]+1):
                            if i not in mapsamps:
                                mapsamps[i] = []
                            mapsamps[i].append(name)
                        opitchdat[name] = note-tune
                        sratedat[name] = samp.sample_rate
                        sampdat[name] = sample
        return {'samples': sampdat, 'sample_rate': sratedat, 'original_pitch': opitchdat,
               'min_note': minmidi, 'max_note': maxmidi, 'sample_map': mapsamps}

    def reconstruct_samples(self, sfpre_dict):
        """Interpolate, combine and resample soundfont samples for each note,
           and load into the `Sampler`.

           Args:
             sfpre_dict (:obj:`dict`): dictionary of data required to load
               soundfont samples in to the `Sampler`, including raw `samples`,
               `sample_rate`, `original_pitch` of the samples, the `min_note`
               and `max_note` in midi values to use the sample, and the
               `sample_map`, assigning each sample to a note.
        
           Return:
             sampdict (:obj:`dict`): output dictionary of mapped notes, with
             values of arrays of sample values at the samplerate of the
             `Generator`.
        """
        minkey = sfpre_dict['min_note']
        maxkey = sfpre_dict['max_note']
        smap = sfpre_dict['sample_map']
        sampdict = {}
        
        for i in range(max(minkey,16), min(maxkey, 115)+1):
            wave_stack = []
            maxlen = 0
            for nme in smap[i]:
                # print((i-sfpre_dict['original_pitch'][nme])/12.)
                semi_shift = pow(2, (i-sfpre_dict['original_pitch'][nme])/12.)
                srate = sfpre_dict['sample_rate'][nme]
                samp = sfpre_dict['samples'][nme]
                vals = utils.resample(semi_shift*srate, self.samprate, samp)
                maxlen = max(maxlen, vals.size)
                wave_stack.append(vals)
            compwave = np.zeros(maxlen, dtype='int16')
            nwave = len(wave_stack)
            for wave in wave_stack:
                compwave[:wave.size] += wave//nwave
            nte = notes.mkey_to_note(i)
            sampdict[nte] = compwave # return notes using sharps
            if nte[1] == '#':
                # if a sharp, also assign flat...
                sampdict[nte.replace('#','b')] = compwave
            # outname = f'../../example_wavs/out_{nte}.wav'
            # write(outname, samprate, compwave)
        return sampdict
            
    def load_samples(self):
        """Load audio samples into the sampler.

        Read audio samples in from a specified directory or via a
        dictionary of filepaths, generate interpolation functions for
        each, and assign them to a named note in scientific pitch notation
        (e.g. :obj:`'A4'`).

        Note:
          Notes are assigned based on a tag in the filename (see :obj:`Sampler`),
          not by analysing the audio itself. If a tuned sample is tagged as the
          wrong note, this will carry over to the sonification. However, this
          allows non-pitched samples to be assigned notes and triggered.
          
        """
        self.samples = {}
        self.samplens = {}
        for note in self.sampdict.keys():
            if isinstance(self.sampdict[note], str):
                rate_in, wavobj = wavfile.read(self.sampdict[note])
                # If it doesn't match the required rate, resample and re-write
                if rate_in != self.samprate:
                    wavobj = utils.resample(rate_in, self.samprate, wavobj)
                # force to mono array, else convert values to float
                if wavobj.ndim > 1:
                    wavdat = np.mean(wavobj.data, axis=1)
                else:
                    wavdat = np.array(wavobj.data, dtype='float64')
            else:
                wavdat = self.sampdict[note].astype('float64')
            # remove DC term 
            dc = wavdat.mean()
            wavdat -= dc
            wavdat /= abs(wavdat).max()
            samps = range(wavdat.size)
            self.samples[note] = interp1d(samps, wavdat,
                                          bounds_error=False,
                                          fill_value = (0.,0.),
                                          assume_sorted=True)
            self.samplens[note] = wavdat.size

    def forward_loopsamp(self, s, start, end):
        """Looping samples forward using indexing

        From a list of samples and start and end loop-points, return a
        new list of samples that index the audio file samples to
        create a forward-looping effect.

        Args:
          s (:obj:`array`-like): Sample indexes for the duration of a
            source's note 
          start (:obj:`int`): Sample index at which to start the loop
          end (:obj:`int`): Sample index at which to end the loop

        Returns:
          s_new (:obj:`array`-like): new sample indices to create a
          forward-looping effect
        """
        delsamp = end-start
        return np.piecewise(s, [s < start, s >= start],
                        [lambda x: x, lambda x: (x-start)%(delsamp) + start])
    def forward_back_loopsamp(self, s, start, end):
        """Looping samples forward-backward alternately using indexing

        From a list of samples and start and end loop-points, return a
        new list of samples that index the audio file samples to
        create a back and forth looping effect.

        Args:
          s (:obj:`array`-like): Sample indexes for the duration of a
            source's note 
          start (:obj:`int`): Sample index at which to start the loop
          end (:obj:`int`): Sample index at which to end the loop

        Returns:
          s_new (:obj:`array`-like): new sample indices to create a
          back and forth looping effect
        
        """
        delsamp = end-start
        return np.piecewise(s, [s < start, s >= start],
                            [lambda x: x,
                             lambda x: end - abs((x-start)%(2*(delsamp)) - (delsamp))])

    def play(self, mapping):
        """ Play the sound for a given source.

        Play a given source and return the sample values for
        combination into the overall sonification.

        Note:
          :obj:`mapping` is a linear dictionary (not nested, as for 
          :meth:`strauss.generator.modify_preset`) where group members
          are indicated using :obj:`'/'` notation
          (e.g. :obj:`{'volume_envelope/A': 0.5, ...`).

        Args:
          mapping (:obj:`dict`): keys and items are generator
            parameter names and their values. This combines all the
            preset mapped parameters, overwritten by any
            :obj:`Source`-mapped parameters (represented as values or
            interpolation functions for static and evolving
            parameters, respectively). This is a linear dictionary
            (not nested, see :meth:`strauss.generator.modify_preset`)
            where group members are indicated using :obj:`'/'`
            notation (e.g. :obj:`{'volume_envelope/A': 0.5, ...`).
        """
        # TO DO: Generator should know samplerate and audbuff
        # TO DO: split this into common and generator-specific functions to minimise code duplication
        samprate = self.samprate
        audbuff = self.audbuff

        params = copy.deepcopy(self.preset)
        utils.linear_to_nested_dict_reassign(mapping, params)
        # for p in self.preset.keys():
        #     if p not in mapping:
        #        mapping[p] = self.preset[p]

        # sample to use
        samplefunc = self.samples[params['note']]
        
        # note length
        if params['note_length'] == 'sample':
            nlength = self.samplens[params['note']]
            params['note_length'] = nlength/samprate
        else:
            nlength = (params['note_length']+params['volume_envelope']['R'])*samprate

        # generator stream (TO DO: attribute of stream?)
        sstream = stream.Stream(nlength/samprate, samprate)
        sstream.get_sampfracs()
        samples = sstream.samples.astype(float)

        pindex  = np.zeros(samples.size)
        if callable(params['pitch_shift']):
            pindex += params['pitch_shift'](sstream.sampfracs)/12.
        elif params['pitch_shift'] != 0:
            pindex += params['pitch_shift']/12.
        if params['pitch_lfo']['use']:
            pindex += self.lfo(samples, sstream.sampfracs, params, 'pitch')/12.
        if np.any(pindex):
            samples = np.cumsum(pow(2., pindex))
        
        # if callable(params['pitch_shift']):
        #     pshift = np.cumsum(params['pitch_shift'](sstream.sampfracs))
        #     samples *= pow(2., pshift/12.)
        # else:
        #     samples *= pow(2., params['pitch_shift']/12.)
        
        # sample looping if specified
        if params['looping'] != 'off':
            startsamp = params['loop_start']*samprate
            endsamp = params['loop_end']*samprate

            # find clean loop points within an audible (< 20Hz) cycle
            startsamp += np.argmin(samplefunc(np.arange(audbuff) + startsamp))
            endsamp += np.argmin(samplefunc(np.arange(audbuff) + endsamp))

            if params['looping'] == 'forwardback':
                samples = forward_back_loopsamp(samples,#sstream.samples,
                                                startsamp,
                                                endsamp)
            elif params['looping'] == 'forward':
                samples = forward_loopsamp(samples,#sstream.samples,
                                           startsamp,
                                           endsamp)
        
                
        # generate stream values
        values = samplefunc(samples)

        # get volume envelope
        env = self.envelope(sstream.samples, params)
        if params['volume_lfo']['use']:
            env *= np.clip(1.-self.lfo(sstream.samples, sstream.sampfracs,
                                       params, 'volume')*0.5, 0, 1)
        # apply volume normalisation or modulation (TO DO: envelope, pre or post filter?)
        sstream.values = values * env * utils.const_or_evo(params['volume'], sstream.sampfracs)
        
        # TO DO: filter envelope (specify as a cutoff array function? or filter twice?)

        # filter stream
        if params['filter'] == "on":
            if hasattr(params['cutoff'], "__iter__"):
                # if static cutoff, use minimum buffer count
                sstream.bufferize(sstream.length/4)
            else:
                # 30 ms buffer (hardcoded for now)
                sstream.bufferize(0.03)
            sstream.filt_sweep(getattr(filters, params['filter_type']),
                               utils.const_or_evo_func(params['cutoff']))
        return sstream    

class Spectralizer(Generator):
    """Spectralizer generator class

    This generator class synthesises sound from a spectrum input
    using an *inverse Fast Fourier Transform* (iFFT) algorithm.
    Defining a minimum and maximum frequency in Hz, input spectrum
    is interpolated between these points such that the output
    audio signal has the requested length. Phases are randomised
    to avoid phase correlations.

    Attributes:
      gtype (:obj:`str`): Generator type 

    Todo:
    	* Add other synthesiser types, aside from additive (e.g. FM,
    	  vector, wavetable)? 
    """

    def __init__(self, params=None, samprate=48000):
        """
        Args:
    	  params (`optional`, :obj:`dict`): any generator parameters
    	    that differ from the generator :obj:`preset`, where keys
            and values are parameters names and values respectively. 
    	  samprate (`optional`, :obj:`int`): the sample rate of
  	    the generated audio in samples per second (Hz)
        """
        # default synth preset
        self.gtype = 'spec'
        self.preset = getattr(presets, self.gtype).load_preset()
        self.preset['ranges'] = getattr(presets, self.gtype).load_ranges() 

        self.eq = utils.Equaliser()
        self.freqwarn = True

        # universal initialisation for generator objects:
        super().__init__(params, samprate)

    def spectrum_to_signal(self, spectrum, phases, new_nlen, mindx, maxdx, interp_type):
        """ Convert the input spectrum into sound signal
        
        Performs the inverse fast fourier transform to produce spectral
        sonification.
        
        Args:
          spectrum (:obj:`ndarray`): Values of the spectrum, ordered
            from high to low frequency
          phases (:obj:`ndarray`): Array of values of `[0,2*numpy.pi]`
            representing the complex number argument
          new_nlen (:obj:`int`): Number of samples needed to enclose
            the output signal.
          mindx (:obj:`int`): Index in total Fourier transform
            represnting the minimum audio frequency
          maxdx (:obj:`int`): Index in total Fourier transform
            represnting the maximum audio frequency
          interp_type (:obj:`str`): Interpolation approach, either
            `"sample"` interpolating between samples, or
            `"preserve_power"` where cumulative power is interpolated
            and then differentiated to avoid missing power.
        """        
        
        if interp_type == "sample":
            ps = np.interp(np.linspace(0,1,maxdx-mindx), np.linspace(0, 1, spectrum.size), spectrum)
        elif interp_type == "preserve_power":
            # we don't renormalise by len(ps) / spectrum.size,
            # as renormalising to peak later anyway.
            ps = np.diff(np.interp(np.linspace(0, 1, maxdx-mindx+1),
                                   np.linspace(0, 1, spectrum.size),
                                   np.cumsum(spectrum)))
           
        empt = np.zeros(new_nlen)
        empt[mindx:maxdx] = ps
        ps = empt
        PS = ps*np.cos(phases) + 1j*ps*np.sin(phases)
        return np.real(ifft(PS))[:new_nlen]
        
    def play(self, mapping):
        """ Play the sound for a given source.

        Play a given source and return the sample values for
        combination into the overall sonification.

        Note:
          :obj:`mapping` is a linear dictionary (not nested, as for 
          :meth:`strauss.generator.modify_preset`) where group members
          are indicated using :obj:`'/'` notation
          (e.g. :obj:`{'volume_envelope/A': 0.5, ...`).

        Args:
          mapping (:obj:`dict`): keys and items are generator
            parameter names and their values. This combines all the
            preset mapped parameters, overwritten by any
            :obj:`Source`-mapped parameters (represented as values or
            interpolation functions for static and evolving
            parameters, respectively). This is a linear dictionary
            (not nested, see :meth:`strauss.generator.modify_preset`)
            where group members are indicated using :obj:`'/'`
            notation (e.g. :obj:`{'volume_envelope/A': 0.5, ...`).
        """
        samprate = self.samprate
        audbuff = self.audbuff

        params = copy.deepcopy(self.preset)
        utils.linear_to_nested_dict_reassign(mapping, params)

        duration = (params['note_length']+params['volume_envelope']['R'])
        nlength = int(duration*samprate)
        # generator stream (attribute of stream?)
        sstream = stream.Stream(nlength/samprate, samprate)
        samples = sstream.samples
        sstream.get_sampfracs()

        spectrum = params['spectrum']
        interp_type = params['interpolation_type']
        
        if np.array(spectrum).ndim == 1:
            # number of discrete frequencies available in ifft between freq. limits 
            discrete_freqs = duration*(params['max_freq']-params['min_freq'])
            
            # how many spectra points fit into the available intermediate frequencies
            spectra_multiples = (discrete_freqs - 1)/(spectrum.size - 1)
            
            # the minimum factor by which to increase the stream length to accomodate spectra in whole number multiples
            if params['fit_spec_multiples']:
                buffer_factor = np.ceil(spectra_multiples)/spectra_multiples
            else:
                buffer_factor = 1
                
            # number of samples to generate including buffer
            new_nlen = int(buffer_factor * nlength)
        
            # the frequency bound indices which the spectrum will be mapped into
            mindx = int(params['min_freq'] * duration * buffer_factor)
            maxdx = int(params['max_freq'] * duration * buffer_factor)

            if params['equal_loudness_normalisation']:
                freqs =  np.linspace(params['min_freq'], params['max_freq'], len(spectrum))
                norm = self.eq.get_relative_loudness_norm(freqs)
                if not self.eq.factor_rms:
                    self.eq.factor_rms = []
                rms1 = np.sqrt(np.mean(spectrum**2))
                spectrum *= norm
                self.eq.factor_rms.append(np.sqrt(np.mean(spectrum**2))/rms1)
                
            # hardcode phase randomisation for now
            phases = 2*np.pi*np.random.random(new_nlen)
            
            # generate stream values
            sstream.values = self.spectrum_to_signal(spectrum, phases, new_nlen, mindx, maxdx, interp_type)[:nlength]
        else:

            if 'time_evo' in params:
                np.diff(params['time_evo'])

            nspec, nwlen  = np.array(spectrum).shape
                
            # buffer for each spectrum
            buffdur = duration / (nspec-1)
            sstream.bufferize(buffdur)

            # indices of IFFT input spectrum corresponding to the nearest desired frequencies
            # NOTE: we don't force the spectrum to reproduce desired frequencies to the accuracy of
            # a few samples in the evolving spectrum case.
            mindx = np.round(params['min_freq'] * buffdur).astype(int)
            maxdx = np.round(params['max_freq'] * buffdur).astype(int)

            if ((maxdx - mindx)*10 < nwlen) and self.freqwarn and params['interpolation_type'] == 'sample':

                basewarn = ("\n\n Spectrum strongly undersampled (by more than a factor 10) while using 'sample' \n"
                            "interpolation type. This could miss spectral features. You could consider: \n"
                            "\t - Changing interpolation type to 'preserve power' (i.e. generator.modify_preset({'interpolation_type':'preserve_power'})) \n")

                fdiff = params['max_freq']- params['min_freq']
                newdur = (nwlen/fdiff) *(nspec-1)
                if newdur < 300: # i.e. 5 minutes
                    # suggest increasing duration if reasonable 
                    basewarn += f"\t - Increase the duration of the sonification (e.g. to > {newdur:.0f}) \n"
                if ((2e4 - 30) * buffdur > nwlen):
                    newdiff = np.log10(nwlen / buffdur)
                    flo = pow(10,0.5*(np.log10(30) + np.log10(2e4)) - 0.5*newdiff)
                    fhi = pow(10,0.5*(np.log10(30) + np.log10(2e4)) + 0.5*newdiff)
                    basewarn += f"\t - Increase the sound frequency range eg ({np.floor(flo):.0f} to {np.ceil(fhi):.0f} Hz)\n"

                warnings.warn(basewarn
                    + f"\t - Rebin spectra more coarsely \n")
                    
                # only warn once per instance
                self.freqwarn = False
            
            # length of buffer and therefore IFFT in this case
            new_nlen = sstream.buffers._nsamp_buff
            
            # hardcode phase randomisation for now
            phases = 2*np.pi*np.random.random(new_nlen)
            
            # iterate through buffers and spectra
            nolap = nspec-1
            buffsize = sstream.buffers._nsamp_buff

            for i in range(nspec):
                # print(buffsize)
                buffs = self.spectrum_to_signal(spectrum[i], phases, new_nlen,
                                                mindx, maxdx, interp_type)
                # print(sstream.buffers._nsamp_buff, buffs.size)
                sstream.buffers.buffs_tile[i] = buffs[:sstream.buffers._nsamp_buff]
                if i == nolap:
                    continue
                # print(sstream.buffers.buffs_tile[i][0], sstream.buffers.buffs_tile[i][-1])
                sstream.buffers.buffs_olap[i][buffsize//2:] = sstream.buffers.buffs_tile[i][:buffsize//2]
                sstream.buffers.buffs_olap[i][:buffsize//2] = sstream.buffers.buffs_tile[i][buffsize//2:]

                if params['regen_phases']:
                    # regenerate randomised phases if doing so
                    phases = 2*np.pi*np.random.random(new_nlen)
                  
            sstream.consolidate_buffers()
            
        sstream.values /= abs(sstream.values).max()
        
        # get volume envelope
        env = self.envelope(sstream.samples, params)
        if params['volume_lfo']['use']:
            env *= np.clip(1.-self.lfo(sstream.samples, sstream.sampfracs,
                                       params, 'volume')*0.5, 0, 1)

        pindex  = np.zeros(samples.size)
        if callable(params['pitch_shift']):
            pindex += params['pitch_shift'](sstream.sampfracs)/12.
        elif params['pitch_shift'] != 0:
            pindex += params['pitch_shift']/12.
        if params['pitch_lfo']['use']:
            pindex += self.lfo(samples, sstream.sampfracs, params, 'pitch')/12.
        if np.any(pindex):
            sampfunc = interp1d(samples, sstream.values,
                                bounds_error=False,
                                fill_value = (0.,0.),
                                assume_sorted=True)
            newsamp = np.cumsum(pow(2., pindex))
            sstream.values = sampfunc(newsamp)

        # apply volume normalisation or modulation (TO DO: envelope, pre or post filter?)
        sstream.values *= utils.const_or_evo(params['volume'], sstream.sampfracs) * env

        # filter stream
        if params['filter'] == "on":
            if hasattr(params['cutoff'], "__iter__"):
                # if static cutoff, use minimum buffer count
                sstream.bufferize(sstream.length/4)
            else:
                # 30 ms buffer (hardcoded for now)
                sstream.bufferize(0.03)
            sstream.filt_sweep(getattr(filters, params['filter_type']),
                               utils.const_or_evo_func(params['cutoff']))
        return sstream    
    
def gen_chord(stream, chordname, rootoctv=3):
    """DEPRECATED CODE:
    generate chord over entire stream given chord name and optional
    octave of root note
    """
    frqs = notes.parse_chord(chordname, rootoctv)
    frqsamp = frqs/stream.samprate 
    for f in frqsamp:
        stream.values += detuned_saw(stream.samples, f)
    
def detuned_saw(samples, freqsamp, oscdets=[1,1.005,0.995]):
    """DEPRECATED CODE: 
    Three oscillator sawtooth wave generator with slight detuning for
    texture
    """
    saw = lambda freqsamp, samp: 1-((samples*(freqsamp)/2) % 2)
    signal = np.zeros(samples.size)
    for det in oscdets:
        freq = freqsamp*det
        signal += saw(freq, samples+freq*np.random.random())
    return signal

def legacy_env(t, dur,a,d,s,r):
    """ DEPRECATED CODE:
    older function for generating envelopes
    """
    att = lambda t: t/a
    dgrad = (1-s)/d
    dec = lambda t: (a-t)*dgrad + 1
    sus = lambda t: s
    funcs = [att, dec, sus]
    conds = [t<a,
             np.logical_and(t>a, t<(d+a)),
             t > (a+d)]
    vol = np.piecewise(np.clip(t, 0, dur), conds, funcs)
    rel = np.clip(np.exp((dur-t)/r),0, 1)
    print(a,d,s,r)
    return vol * rel
    
if __name__ == "__main__":
    # test volume envelope
    t = np.linspace(0.,11,500)
    dur = 9
    a = 1.4
    d = 2
    s = 0.7
    r = 1
    env = legacy_env(t, dur,a,d,s,r)
    plt.plot(t, env)
    plt.show()



""" The :obj:`utilities` submodule: useful functions for ``strauss``

This submodule is for useful utility functions used by other
``strauss`` modules. Generally these are not intended for direct
use by the user.
"""
from functools import reduce
import operator
import numpy as np
from scipy.interpolate import interp1d
from contextlib import contextmanager,redirect_stderr,redirect_stdout
from os import devnull
from io import StringIO 
import sys
from pathlib import Path

# Some utility classes (these may graduate to somewhere else eventually)

class NoSoundDevice:
    """
    Drop-in replacement for sounddevice module if not working,
    so we can still use other functionality.

    Attributes:
      err (:obj:`Exception`): Error message from trying to import
        sounddevice
    """
    def __init__(self, err):
        self.err = err
    def play(self, *args, **kwargs):
        """Dummy function replacing `sounddevice.play` when unavailable.

        Args:
          *args: arguments (ignored)
          **kwargs: keyword-only arguments (ignored)
        """
        raise self.err

class Equaliser:
    def __init__(self):

        self.factor_rms = None
        parpath = Path(f"{Path(__file__).parent}","data","params.csv")
        # Read in parameters for ISO 226:2024 standard.
        pars = np.genfromtxt(parpath, delimiter=',', names=True)
        self.parfuncs = {}
        for c in pars.dtype.names:
            if c == 'freq':
                continue
            self.parfuncs[c] = interp1d(np.log10(pars['freq']),
                                        pars[c],
                                        fill_value='extrapolate')
            
    def get_relative_loudness_norm(self, freq, phon=70.):
        """
        Relative normalisation of sound frequencies
        To compensate for pereceptual loudness, following
        the ISO 226:2024 standard.

        Args:
          freq (:obj:`array-like`) audio frequencies in Hz
          phon (:obj:`float`) listening level for a 1 kHz
            note 

        Returns:
          rnorm (:obj:`array-like`) volume normalisation
            for spectra
        """
        lfreq = np.log10(freq)
        L_U = self.parfuncs['L_U'](lfreq)
        alpha_f = self.parfuncs['alpha_f'](lfreq)
        T_f = self.parfuncs['T_f'](lfreq)

        A = pow(4e-10, 0.3 - alpha_f)
        B = pow(10., (phon)*3e-2) - pow(10, 7.2e-2)
        C = pow(10., alpha_f * 0.1*(T_f + L_U))

        L_f = 10*np.log10(A*B + C)/alpha_f - L_U
        norm = pow(10., (L_f - phon)/20)
        rnorm = norm/norm.max()
        return rnorm

    
# a load of utility functions used by STRAUSS

def nested_dict_reassign(fromdict, todict):
    """
    Recurse through dictionaries and sub-dictionaries in
    `fromdict` and reassign equivalent values in `todict`

    Args
     fromdict (:obj:`dict`): Dictionary containing values
      to assign
     todict (:obj:`dict`): Dictionary containing values
      to be reassigned
    """
    for k, v in fromdict.items():
        if isinstance(v, dict):
            # recurse through nested dictionaries
            nested_dict_reassign(v, todict[k])
        else:
            # reassign todict value
            todict[k] = v

def nested_dict_fill(fromdict, todict):
    """
    Recurse through dictionaries and sub-dictionaries in
    `fromdict` and assign to any entries missing from
    `todict` 
    
    Args:
     fromdict (:obj:`dict`): Dictionary containing values
      to assign
    todict (:obj:`dict`): Dictionary containing values to
      be reassigned
    """
    for k, v in fromdict.items():
        if k not in todict:
            # assign todict value
            todict[k] = v
        elif isinstance(v, dict):
            # recurse through nested dictionaries
            nested_dict_fill(todict[k], v)
            
def nested_dict_idx_reassign(fromdict, todict, idx):
    """
    Recurse through dictionaries and sub-dictionaries of
    iterables in `fromdict` and index value idx to assign
    or replact value in todict
    
    Args:
     fromdict (:obj:`dict`): Dictionary containing values
      to assign
     todict (:obj:`dict`): Dictionary containing values
      to be reassigned
     idx (:obj:`dict`): Index value for retrieving value
      from iterables
    """
    for k, v in fromdict.items():
        if isinstance(v, dict):
            # recurse through nested dictionaries
            nested_dict_idx_reassign(todict[k], v, idx)
        else:
            # reassign todict value
            todict[k] = v[idx]

def reassign_nested_item_from_keypath(dictionary, keypath, value):
    """
    Reassign item in a nested dictionary to value using keypath syntax,
    to traverse multiple dictionaries

    Args:
     dictionary (:obj:`dict`): dict object to reassign values within
     keypath (:obj:`str`): Using filepath syntax on given OS to
       traverse dictionary, i.e 'a/b/c' ('a\\b\\c') corresponds to
       dict['a']['b']['c'] on Unix (Windows).
     value: value to reassign dictionary value with
    """
    p = Path(keypath)
    keylist = list(p.parts)
    get_item = lambda d, kl: reduce(operator.getitem, kl, d)
    get_item(dictionary, keylist[:-1])[keylist[-1]] = value
            
def linear_to_nested_dict_reassign(fromdict, todict):
    """
    Iterate through a linear dictionary to reassign nested values
    using keypaths (d1['a/b/c'] -> d2['a']['b']['c'], d1['a']->d2['a'])

    Args:
     fromdict (:obj:`dict`):
      Dictionary containing values to assign
     todict (:obj:`dict`):
      Dictionary containing values to be reassigned    
    """
    for k, v in fromdict.items():
        reassign_nested_item_from_keypath(todict, k, v)
            
def const_or_evo_func(x):
    """
    If x is callable, return x, else provide a function that just
    returns x

    Args:
      x: input value, either a numerical value or a
      function
    """
    if callable(x):
        return x
    else:
        return lambda y: y*0 + x

def const_or_evo(x,t):
    """
    If x is callable, return x(t), else return x

    Args:
      x: input value, either a numerical value or a
      function
      t (numerical): values to evaluate x function
    """
    if callable(x):
        return x(t)
    else:
        return x

def rescale_values(x, oldlims, newlims):
    """
    Rescale x values defined by limits oldlims to new limits newlims

    Args:
      x (array-like): Array of input values
      oldlims (:obj:`tuple`): tuple representing the original limits
      of `x` (low, high)
      newlims (:obj:`tuple`): tuple representing the new limits

    Returns:
      x_rs (array-like): Rescaled array
    """
    olo, ohi = oldlims
    nlo, nhi = newlims
    descale = np.clip((x - olo) / (ohi-olo), 0 , 1)
    return (nhi-nlo)*descale + nlo
    
def resample(rate_in, samprate, wavobj):
    """
    Resample audio from original samplerate to required samplerate

    Args:
      rate_in (:obj:`int`) sample rate of input wave object
      samprate (:obj:`int`) desired sample rate for output
      wavobj (:obj:`tuple`) sample rate, sample array tuple, output
      by `scipy.io.wavfile` function

    Returns:
      new_wavobj (:obj:`tuple`) as `wavobj`, with new sample rate
      and resampled sample values
    """
    duration = wavobj.shape[0] / rate_in

    time_old  = np.linspace(0, duration, wavobj.shape[0])
    time_new  = np.linspace(0, duration,
                            int(wavobj.shape[0] * samprate / rate_in))

    interpolator = interp1d(time_old, wavobj.T)
    new_wavobj = np.round(interpolator(time_new).T).astype(wavobj.dtype)
    return(new_wavobj)

@contextmanager
def suppress_stdout_stderr():
    """
    A context manager that redirects stdout and stderr to devnull
    """
    with open(devnull, 'w') as fnull:
        with redirect_stderr(fnull) as err, redirect_stdout(fnull) as out:
            yield (err, out)

def get_supported_coqui_voices():
    """Show supported Coqui-AI TTS voices
    
    """
    voices = []

    jenny = {"id": Path('tts_models', 'en', 'jenny', 'jenny'),
             "name": 'jenny',
             "languages": ['en_GB'],
             "age": None}
    
    voices.append(jenny)

    vits = {"id": Path('tts_models', '{iso_code}', 'fairseq', 'vits'),
            "name": 'fairseq-{iso_code}',
            "languages": ['many (https://dl.fbaipublicfiles.com/mms/tts/all-tts-languages.html for iso codes)'],
            "age": None}

    voices.append(vits)

    vits_ita = {"id": Path('tts_models', 'it', 'mai_female', 'glow-tts'),
                "name": 'mai',
                "languages": ['it_IT'],
                "age": None}

    voices.append(vits_ita)

    thorsten = {"id": Path('tts_models', 'de', 'thorsten', 'vits'),
                "name": 'thorsten',
                "languages": ['de_DE'],
                "age": None}

    voices.append(thorsten)

    return voices

    
class Capturing(list):
    """
    Context manager for handling stdout (see https://stackoverflow.com/a/16571630)
    """
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self
    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio    # free up some memory
        sys.stdout = self._stdout



""" :obj:`channels` submodule representing the output audio channels.

This submodule defines objects relevant to the output audio channels,
including the :obj:`audio_channels` which defines arrays of microphone
objects that are channeled to different speakers in the sonification
output.

"""

import numpy as np
import matplotlib.pyplot as plt
import warnings
import sys
from scipy.special import lpmv, factorial
class mic:
    """Microphone / sound detector object

    This class represents a microphone, or sound-detector,
    corresponding to a particular output channel for audio
    spatialisation in the sonicfication.

    Args: 
      azimuth (:obj:`float`): Angular position of the microphone on the
    	horizontal plane, from 0 to 2pi with 0.5 pi being to the left
    	and 1.5 pi being to the right. In the special case of ambisonic,
    	this is instead an index corresponding to the *Ambisonic Channel
	Number* (ACN)
      mic_type (:obj:`str`): Type of microphone, choose from
    	:obj:`"directional"` (collects using a cardioid antenna pattern),
    	:obj:`"omni"` (collects sound from all directions equally) and
    	:obj:`"mute"` (collects no sound, useful for e.g. muting
    	auxillary channels)
      label (:obj:`str`): A label for the mic
      channel (:obj:`int`) The index of the channel, corresponding to
    	channel ordering in the output (starting from 0 e.g. stereo
    	L=0, R=1) 

    Raises:
      Exception: if mic_type not in allowed options
    """
    def __init__(self, azimuth, mic_type="directional", label="C", channel=1,  order=None, degree=None):
        self.azimuth = azimuth
        self.mic_type = mic_type
        self.label = label
        self.channel = channel
        
        if mic_type == "directional":
            self.antenna = lambda a, b=0.5*np.pi: 0.5*(1+np.cos(a-azimuth)*np.sin(b))
        elif mic_type == "omni":
            self.antenna = lambda a, b=0.5*np.pi: a**0.
        elif mic_type == "mute":
            self.antenna = lambda a, b=0.5*np.pi: a*0.
        elif mic_type == "ambisonic":
            self.antenna = self._ambisonic_antenna(acn=azimuth)
        else:
            raise Exception(f"Mic type \"{mic_type}\" unknown.")
        
    def _ambisonic_antenna(self, acn):        
        # get order and degree of spherical harmonic from ACN using ambiX standard
        # (see iem.kug.ac.at/fileadmin/media/iem/projects/2011/
        # ambisonics11_nachbar_zotter_sontacchi_deleflie.pdf, footnote 5)
        # with modified l expression protecting against zero division in acn == 0 case.
        l = int(acn**0.5)
        m = acn - l*(l+1)
        mabs = abs(m)

        # normalise using SN3D standard (see iem.kug.ac.at/fileadmin/media/
        # iem/projects/2011/ambisonics11_nachbar_zotter_sontacchi_deleflie.pdf,
        # equation 3)
        fctrl = factorial
        normSN3D = np.sqrt((2-(0**mabs)/4*np.pi) * fctrl(l-mabs)/fctrl(l+mabs))

        # trig function to use, dependent on sense of m (see ref eq 2)
        if m < 0:
            tfunc = np.sin
        if m >= 0:
            tfunc = np.cos
        
        # return lambda function, indexing correct spherical harmonic and
        # normalising to  SN3D normalisation (ref equation 2).
        # Note: additonal (-1)^mabs term needed to match ambiX given differing
        # assoc. Legendre polynom. definitions between ambiX and numpy. 
        return lambda a, b=0.5*np.pi: normSN3D * pow(-1,mabs) * lpmv(mabs, l, np.cos(b)) * tfunc(mabs*a)
        
    
class audio_channels:
    """Representing output audio channels.

    Data object representing the output channels of the sonification
    for preset common audio setups, or a custom setup. Stores an array
    of :obj:`mic` objects for each output channel.
    
    Args:
      setup (:obj:`str`): Type of audio setup. Supported options are
        :obj:`"mono"`, :obj:`"stereo"`, :obj:`"5.1"` and
        :obj:`"7.1"`, or :obj:`"custom"`. 
      custom_setup (:obj:`dict`): Dictionary defining a customised
    	audio setup, containing keys for :obj:`"azimuths"`, :obj:`"types"`
    	and :obj:`"labels"`, containing lists parameterising the first
    	three arguments of the :class:`mic` object, respectively
    	in the order of their channel index. Also optionally an forder
    	list to unscramble any channel order scrambling done by ffmpeg
    	processing in the sonification sae routines (this may need to
    	be found empirically, awaiting better multichannel save
    	routine).

    Raises:
	Exception: If custom requested but no parameters provided, or
    	  custom parameters provided without requesting a custom setup.

    """

    def __init__(self, setup="stereo", custom_setup={}):

        ##############################################
        # Channel properties for preset audio setups
        ##############################################
        
        # mic angles in radians
        mono_azimuths = [0.]
        stereo_azimuths = [0.5*np.pi, 1.5*np.pi]
        fivepoint_azimuths = [1./3*np.pi, 5./3*np.pi,
                            0., 0.,
                            2./3*np.pi, 4./3*np.pi]
        sevenpoint_azimuths = fivepoint_azimuths + stereo_azimuths

        # mic type, either omni(-directional) directional,
        # or muted (mute channels not used for spatialisation)

        mono_types = ["omni"]
        stereo_types = ["directional"] * 2
        fivepoint_types = ["directional"] * 2  + \
                          ["mute"] * 2 + \
                          ["directional"] * 2
        fivepoint_types = ["directional"] * 2  + \
                          ["mute"] * 2 + \
                          ["directional"] * 2
        sevenpoint_types = ["directional"] * 2  + \
                           ["mute"] * 2 + \
                           ["directional"] * 4
        
        # mic labels corresponding to each speaker
        mono_labels = ['C']
        stereo_labels = ['L', 'R']
        fivepoint_labels = ['FL', 'FR', 
                            'FC', 'LF',
                            'SL', 'SR']
        sevenpoint_labels = ['FL', 'FR', 
                             'FC', 'LF',
                             'SL', 'SR',
                             'AL', 'AR']

        #ffmpeg channel orders to correctly save combined files
        mono_forder = [0]
        stereo_forder = [0,1]
        fivepoint_forder = [1,2,0,3,4,5]
        sevenpoint_forder = [1,2,0,3,4,5,6,7]
        
        self.setup = setup

        if custom_setup and (setup != "custom"):
            warnings.warn("custom_setup variable non-empty, but not using custom setup. " \
                          "Did you mean to set setup=\"custom\"?")
        if  (not bool(custom_setup)) and (setup == "custom"):
            raise Exception("Custom setup requested but custom_setup parameters empty. " \
                            "Please provide setup dictionary to custom_setup")

            
        if setup == "mono":
            self.setup_channels(mono_azimuths, mono_types, mono_labels)
            self.forder = mono_forder
        elif setup == "stereo":
            self.setup_channels(stereo_azimuths, stereo_types, stereo_labels)
            self.forder = stereo_forder
        elif setup == "5.1":
            self.setup_channels(fivepoint_azimuths, fivepoint_types, fivepoint_labels)
            self.forder = fivepoint_forder
        elif setup == "7.1":
            self.setup_channels(sevenpoint_azimuths, sevenpoint_types, sevenpoint_labels)
            self.forder = sevenpoint_forder
        elif setup[:-1] == 'ambiX':
            # i.e. ambiX3 => 3rd order ambisonics.
            nchan = np.sum(2*np.arange(int(setup[-1])+1).astype(int) + 1)
            labfunc = lambda a, b: str(a)+str(b)
            self.setup_channels(np.arange(nchan, dtype='int'),
                                ['ambisonic']*nchan,
                                list(map(labfunc, ['C']*nchan, range(nchan))))
        elif setup == "custom":
            self.setup_channels(custom_setup['azimuths'],
                                custom_setup['types'],
                                custom_setup['labels'])
            if 'forder' in custom_setup:
                self.forder = self.custom_setup['forder']
            else:
                self.forder = 'unknown'
        else:
            raise Exception(f"setup \"{setup}\" not understood")
            

    def setup_channels(self, azimuths, types, labels, orders=None, degrees=None):
        """initialise audio channel setup for lists of properties

        Subroutine for setting up the audio_channels as arrays of
        :obj:`mic` objects, setting the :obj:`self.mics` list
        attribute to the :obj:`audio_channels`

        Args:
          azimuths (:obj:`list(float)`): list of :obj:`azimuth` values for
        	:obj:`mic` object
          types (:obj:`list(float)`): list of :obj:`mic_types` values
          	for :obj:`mic` object.
          labels (:obj:`list(float)`): list of :obj:`label` values for
          	:obj:`mic` object

        """
        self.azimuths = azimuths
        self.types = types
        self.labels = labels
        self.Nmics = len(azimuths)

        # Note: channel ordering important, sets output channel number
        self.channels = range(1, self.Nmics+1)
        
        self.mics = []
        
        for i in range(self.Nmics):
            microphone = mic(azimuths[i], types[i], labels[i], self.channels[i], orders, degrees)
            self.mics.append(microphone)

    def plot_antenna(self):
        """Plot antennae patterns for chosen audio setup

        Make a :obj:`matplotlib` figure object, representing a radial
        plot demonstrating the antennae patterns of each channel

        Returns:
          fig (:obj:`matplotlib.pyplot.figure`): figure object that
          can be shown or saved using the standard :obj:`matplotlib`
          routines. 
        """

        plot_azimuths = np.linspace(0, 2*np.pi, 100)
        normalised_volume = 0.*plot_azimuths

        # setup axes
        fig = plt.figure(figsize=(8,8))
        ax = fig.subplots(subplot_kw={'projection': 'polar'})
        shift = 0.

        for i in range(self.Nmics):
            labelpos = 1.2
            microphone = self.mics[i]
            antenna = microphone.antenna(plot_azimuths)
            normalised_volume += antenna
            p = ax.plot(plot_azimuths, np.clip(antenna,0, np.inf))
            p2 = ax.plot(plot_azimuths, np.clip(-antenna,0, np.inf),
                         c=p[0].get_color(), ls =':')
            if np.all(microphone.antenna(plot_azimuths) == 0.):
                labelpos += shift
                shift -= 0.15
            ax.scatter(microphone.azimuth, labelpos)
            ax.annotate(microphone.label+f" (ch. {microphone.channel})",
                         (microphone.azimuth, labelpos),
                         (0, 10), textcoords='offset points', ha='center')
        normalised_volume /= normalised_volume.max()
        plt.plot(plot_azimuths, normalised_volume, c='k', ls ='--', label='net amplitude')
        
        # configure axes
        ax = plt.gca()
        ax.set_theta_zero_location("N")
        ax.set_yticklabels([])
        ax.set_xlabel('azimuth')
        ax.set_ylim(0,1.4)
        ax.legend(frameon=0)
        ax.grid(True)
        ax.set_title(f"Audio setup: {self.setup}")

        # return figure object to manipulate, show or save
        return fig
            

        
if __name__ == "__main__":
    # can use setup values, mono, stereo, 5.1, 7.1 or custom
    setup = "5.1"
    if len(sys.argv) > 1:
        # read setup type from command line argument
        setup = sys.argv[1]
    ac = audio_channels(setup=setup)
    fig = ac.plot_antenna()
    plt.show()



"""STRAUSS (Sonification Tools and Resources for Analysis Using Sound Synthesis)

This module provides a toolkit for *sonification*, i.e. the representation of data using sound."""

# relative imports of submodules
from . import channels
from . import filters
from . import generator
from . import notes
from . import score
from . import sonification
from . import sources
from . import stream
from . import presets

__version__ = "1.0.1"



"""The :obj:`notes` submodule: translating musical note representations

This submodule contains functions for translating between different
representations of musical notes or musical chords, and representative
sound frequencies and MIDI notes.

Attributes:
  tuneC0 (:obj:`float`): The frequency in Hz of the ``C0`` musical
    note
  notecount (:obj:`int`): Semitone offset above C in an octave
  notesharps (:obj:`list`): Names of musical notes using sharp notation
  noteflats (:obj:`list`): Names of musical notes using flat notation.
  semitone_dict (:obj:`dict`): Dictionary of note names to semitone
    offsets above C.
"""

import numpy as np
import pychord as chrd
import re

# Tune note system to A440 standard
# Equal temperement

tuneC0 = 440 * pow(2, (-9./12)-4) 

notecount = np.arange(12)
notesharps  = ["C","C#","D","D#","E","F","F#","G","G#","A","A#","B"]
noteflats = ["C","Db","D","Eb","E","F","Gb","G","Ab","A","Bb","B"]
semitone_dict = {**dict(zip(notesharps, notecount)),
                 **dict(zip(noteflats, notecount))}

def parse_note(notename):
    """ 
    Takes scientific pitch name and returns frequency in Hz.
    Flat and sharp values supported. Assumes equal temperament
    and A4 = 440 Hz tuning (ISO 16)

    Args:
      notename (:obj:`str`): scientific pitch name, in format
        <note><octave>, e.g. 'Ab4', 'E3' or 'F#2'

    Returns:
      out (numerical): Frequency of note in Hertz
    """
    nsplit = re.findall("(\D+|\d+)", notename)
    semi = semitone_dict[nsplit[0]]/12.
    octv  = int(nsplit[1])
    return tuneC0*pow(2.,semi+octv)

def parse_chord(chordname, rootoct=3):
    """
    Takes name of a chord and root octave to generate a valid
    chord voicing as an array of frequencies in Hz, using the
    `pychord` library

    Args:
      chordname (:obj:`str`): Standard chord name, e.g. `'A7'`
        or `'Dm7add9'` etc.
      rootoct (:obj:`int`): Octave number

    Returns:
      out (:obj:`ndarray`) array of frequencies constituting
        chord
    """
    chord = chrd.Chord(chordname)
    notes = chord.components_with_pitch(rootoct)
    frqs = []
    for n in notes:
        frqs.append(parse_note(n))
    return np.array(frqs)

def chord_notes(chordname, rootoct=3):
    """
    Takes name of a chord and root octave to generate a valid
    chord voicing as a list of note names, using the `pychord`
    library

    Args:
      chordname (:obj:`str`): Standard chord name, e.g. `'A7'`
        or `'Dm7add9'` etc.
      rootoct (:obj:`int`): Octave number

    Returns:
      out (:obj:`list`): list of note names constituting chord
    """
    chord = chrd.Chord(chordname)
    notes = chord.components_with_pitch(int(rootoct))
    return notes

def mkey_to_note(val):
    """
    Take MIDI key value and return the note name in scientific
    notation

    Args:
      val (:obj:`int`): MIDI key value

    Returns:
      out (:obj:`str`): scientific pitch name, in format
        `<note><octave>`, e.g. `'E3'` or `'F#2'`
    """
    from strauss.notes import notesharps
    octv = val // 12 - 1
    semi = val % 12
    return f'{notesharps[semi]}{octv}'



"""The :obj:`filters` submodule: containing audio filter functions

These are audio filters that can be applied to the audio signal in
frequency space to attenuate (filter out) frequencies. These can be
applied to individual ``Buffers`` as an evolvable parameter.

Todo:
  * Support More Filter Types
  * Implement resonance or `'Q'` variation
"""
import numpy as np
import scipy.signal as sig

def LPF1(data, cutoff, q, order=5):
    """
    Low-pass filter data array given cutoff, q and LPF order

    Args:
      data (array-like): Array containing signal for filtering
      cutoff (:obj:`float`): Cutoff frequency
      q (:obj:`float`): Filter quality-factor or 'Q' value
      order (:obj:`int`): polynomial order of filter function

    Return
      y (array-like): Filtered array for output
    """
    b, a = sig.butter(order, cutoff, btype='low', analog=False)
    y = sig.lfilter(b, a, data)
    return y

def HPF1(data, cutoff, q, order=5):
    """
    High-pass filter data array given cutoff, q and HPF order

    Args:
      data (array-like): Array containing signal for filtering
      cutoff (:obj:`float`): Cutoff frequency
      q (:obj:`float`): Filter quality-factor or 'Q' value
      order (:obj:`int`): polynomial order of filter function

    Return
      y (array-like): Filtered array for output
    """
    b, a = sig.butter(order, cutoff, btype='high', analog=False)
    y = sig.lfilter(b, a, data)
    return y




import yaml
from pathlib import Path

p = Path(__file__)
thisdir = p.parent

def read_yaml(filename):
    with filename.open(mode='r') as fdata:
        try:
            yamldict = yaml.safe_load(fdata)
        except yaml.YAMLError as err:
            print(err)
    return yamldict

def load_ranges(name="default"):
    filename = Path(f"{thisdir}","ranges", f"{name}.yml")
    return read_yaml(filename)

def load_preset(name="default"):
    if Path(name).name == Path(name):
        # if open user directly
        filename = Path(f"{name}.yml")
    else:
        # else load built-in preset of that name
        filename = Path(f"{thisdir}", f"{name}.yml")
    return read_yaml(filename)

def preset_details(name='*'):
    pres = sorted(Path(f"{thisdir}").glob(f"*{name.lower()}*.yml"))
    for p in pres:
        with p.open(mode='r') as fdata:
            try:
                d = yaml.safe_load(fdata)
                print(f"\033[1m{d['name']}:\033[0m\n{d['description']}\n")
            except:
                pass



"""The :obj:`tts_caption` submodule: tool for generating spoken captions

This uses text-to-speech via the the ``TTS`` module to allow captions
represented as strings to be converted to spoken audio to precede the
sonification.
"""

from scipy.io import wavfile
from scipy.interpolate import interp1d
import numpy as np
import strauss.utilities as utils
import re
import ffmpeg as ff
import os
import warnings
from pathlib import Path

default_tts_voice = None

class NoTTSAPI(Exception):
    # except when no API key is found for coqui-TTS module
    pass
try:
    from TTS.api import TTS
    ttsMode = 'coqui-tts'
    default_tts_voice = Path('tts_models','en','jenny', 'jenny')
    supported_voices = utils.get_supported_coqui_voices()    
except (OSError, ModuleNotFoundError, NoTTSAPI) as sderr:
    try:
      import pyttsx3
      ttsMode = 'pyttsx3'
      class TTS:
          def __init__(*args, **kwargs):
              pass
          def list_models(self):
              return getVoices(True)
      warnings.warn("Default TTS module coqui not found, using pyttsx3 instead. Note this is platform \n"
                    "dependent and can be problematic for linux-based systems (using the espeak engine)")
      default_tts_voice = {} # i.e. system default TTS (if exists)
    except (OSError, ModuleNotFoundError) as sderr:
      ttsMode = 'None'
      # print('No supported text-to-speech packages have been found.')
      def TTS(*args, **kwargs):
          raise TTSIsNotSupported("strauss has not been installed with text-to-speech support. \n"
                "This is not installed by default, due to some specific module requirements of the TTS module.\n"
                "Reinstalling strauss with 'pip install strauss[AI-TTS]' will give you access to this function\n"
                "If you run into issues with the TTS package, you can also install pyttsx3. Currently the most\n" 
                "compatible version is not published on PyPI, but you can install from the test repo with \n"
                "'pip install --no-cache-dir --extra-index-url https://test.pypi.org/simple/ pyttsx3==2.99'")

class TTSIsNotSupported(Exception):
    pass

def get_ttsMode():
   return ttsMode

def getVoices(info=False):
  '''Get available voices for text-to-speech.

  When info=True, this prints out information
  for each voice option.

    Args:
      info (:obj:`bool`): Print out voice information when True, 
      by default False
      voices (:obj:`list`): List of ``pyttsx3.voice.Voice`` objects
      or ``dict`` objects.
 
  '''
  if ttsMode == 'pyttsx3':
      engine = pyttsx3.init()
      voices = engine.getProperty('voices')
      getter = vars
  elif ttsMode == 'coqui-tts':
      voices = supported_voices
      getter = dict
  else:
      getter = dict
      voices = [{"voices": "None"}]
  if info==True:
      print('Text-to-speech voice options')
      for ind in range(len(voices)):
          voiceProps = getter(voices[ind])
          print('\nVoice index:', ind)
          for key in voiceProps.keys():
              print('{}: {}'.format(key, voiceProps[key]))      
  return voices      

def render_caption(caption, samprate, model, caption_path):
    '''The render_caption function generates an audio caption from text input
    and writes it as a wav file. If the sample rate of the model is not equal 
    to that passed from sonification.py, it resamples to the correct rate and
    re-writes the file. 
    
    If Coqui-AI is installed, text from user input is converted with text-to-
    speech software from Coqui-AI - https://pypi.org/project/TTS/ . 
    You can view publicly available voice models with 'TTS.list_models()'

    If Coqui-AI is not installed but pyttsx3 (https://pypi.org/project/pyttsx3/)
    is installed, text from user input is converted offline using pyttsx3.

    Note:
    STRAUSS checks if Coqui-AI is available. If it is, ``ttsMode`` is set to
    ``coqui-ai``. If it is unavailable, STRAUSS checks whether pyttsx3 is 
    available. If it is, ``ttsMode`` is set to ``pyttsx3``.

    Args:
      caption (:obj:`str`): script to be spoken by the TTS voice
      samprate (:obj:`int`): samples per second
      model (:obj:`str` for Coqui-AI; :obj:`dict` for pyttsx3): for Coqui-AI: 
        valid name of TTS voice from the underlying TTS module; for pyttsx3:
        dictionary with keys of 'rate' (percent of speed, signed int16),
        'volume' (float from 0 to 1), and/or 'voice' (the voice 'id' that can
        be chosen from the list given by the TTS.list_models() function).
      caption_path (:obj:`str`): filepath for spoken caption output
    '''

    # TODO: allow uniform indexing and/or language querying approaches for more consistency between tts modes...
    if ttsMode == 'coqui-tts':

      # TODO: do this better with logging. We can filter TTS function output, e.g. alert to downloading models...
      print('Rendering caption (this can take a while if the caption is long, or if the TTS model needs downloading)...')
      
      # capture stdout from the talkative TTS module
      with utils.Capturing() as output:
          # Load in the tts model
          tts = TTS(str(model), progress_bar=False, gpu=False)

          # render to speech, and write as a wav file (allow )
          tts.tts_to_file(text=caption, file_path=caption_path)
          
    elif ttsMode == 'pyttsx3':
        
      # Setup voice model for pyttsx3
      engine = pyttsx3.init() # initialize object

      # check what model info was set; if none were
      # specified, use defaults
      for key in ['rate','volume','voice']:
          if key in model.keys():
              engine.setProperty(key, model[key])
          else:
              pass

      engine.save_to_file(caption, caption_path, name='caption')
      # note the current PyPI release ()
      engine.runAndWait()
      
    else:
       # initialise dummy TTS class to raise error.
       TTS()
          
    # Read the file back in to check the sample rate
    try:
        # Try to read in directly...
        rate_in, wavobj = wavfile.read(caption_path)
    except:
        # ...but pttsx3 TTS can produce audio files incompatable
        # with scipy - convert to standard WAV using ffmpeg
        cpre = caption_path.split('.')[0] + '_pre.wav'
        os.rename(caption_path, cpre)
        ff.input(cpre).output(caption_path).run(quiet=1)
        rate_in, wavobj = wavfile.read(caption_path)
        
    # If it doesn't match the required rate, resample and re-write
    if rate_in != samprate:
        new_wavobj = utils.resample(rate_in, samprate, wavobj)
        wavfile.write(caption_path, samprate, new_wavobj)





""" The :obj:`sources` submodule: representing data as sound sources.

This submodule deals with the mapping of input datasets to the parameters
controlling sound in the eventual sonification.   

Attributes:
   mappable (:obj:`list(str)`): List of strings indicating possible
	sonification parameters to which data can be mapped.
   evolvable (:obj:`list(str)`): List of strings indicating the subset of
	`mappable` parameters that can be evolved continuosly for an
	individual Source.
   param_limits (:obj:`list(tuple)`): List of tuples indicating the default
	numerical ranges bounding corresponding mappable parameter
	(e.g. 0-1 for volume).
   param_lim_dict (:obj:`dict`): Dictionary combining `mappable` (keys) and 
	`param_limits` (items).

Todo:
    * Store mappable, evolvable and parameter ranges in YAML files (cleaner). 
    * Specialised Event and Object child classes (eg. spectralisation).
"""

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt
from .utilities import rescale_values 
import warnings

mappable = ['polar',
            'azimuth',
            'theta',
            'phi',
            'volume',
            'pitch',
            'time',
            'cutoff',
            'time_evo',
            'spectrum',
            'pitch_shift',
            'volume_envelope/A',
            'volume_envelope/D',
            'volume_envelope/S',
            'volume_envelope/R',
            'volume_lfo/freq',
            'volume_lfo/freq_shift',
            'volume_lfo/amount',
            'pitch_lfo/freq',
            'pitch_lfo/freq_shift',
            'pitch_lfo/amount']     

evolvable = ['polar',
             'azimuth',
             'theta',
             'phi',
             'volume',
             'cutoff',
             'time_evo',
             'pitch_shift',
             'volume_lfo/freq_shift',
             'volume_lfo/amount',
             'pitch_lfo/freq_shift',
             'pitch_lfo/amount']
param_limits = [(0,1),#np.pi),
                (0,1),#2*np.pi),
                (0,1),#np.pi),
                (0,1),#2*np.pi),
                (0,1),
                (0,1),
                (0,1),
                (0,1),
                (0,1),
                (0,1),
                (0,24),
                (1e-2, 10),
                (1e-2, 10),
                (0,1),
                (1e-2, 10),
                (1,12),
                (0,3),
                (0,1),
                (1,12),
                (0,3),
                (0,2)]     

param_lim_dict = dict(zip(mappable, param_limits))

class Source:
    """ Generic source class defining common methods/attributes
    
    `Source` and its child classes represent the input data, and its
    mapping to sonification parameters.

    Note:
	`Source` isn't used directly, instead use child classes
    	`Events` or `Objects`.

    Attributes:
      mapped_quantities (:obj:`list(str)`): The subset of parameters to
        which data will be mapped.
      raw_mapping (:obj:`dict`): Housing the input mapped parameters
        and data, with keys corresponding to :obj:`mapped_quantities`.
      mapping (:obj:`dict`): processed mapping :obj:`dict` rescaled
        to parameter ranges, or interpolation funtions for evolving
        parameters.
    
    Raises:
    	UnrecognisedProperty: if `mapped_quantities` entry not in `mappable`.
    """
    def __init__(self, mapped_quantities):
        """
        Args:
    	  mapped_quantities (:obj:`list(str)`): The subset of parameters to
    	    which data will be mapped.
        """
        # check these are all mappable parameters

        
        for q in mapped_quantities:
            if q not in mappable:
                raise UnrecognisedProperty(
                    f"Property \"{q}\" is not recognised")

        if ('theta' in mapped_quantities) and ('polar' in mapped_quantities):
            raise Exception(
                "\"theta\" and \"polar\" cannot be combined as " \
                "these represent the same quantity: \"theta\" and " \
                "\"phi\" are deprecated and will be replaced with \"polar\"" \
                " and \"azimuth\" in a future version.")

        if ('phi' in mapped_quantities) and ('azimuth' in mapped_quantities):
            raise Exception(
                "\"phi\" and \"azimuth\" cannot be combined as " \
                "these represent the same quantity: \"theta\" and " \
                "\"phi\" are deprecated and will be replaced with \"polar\"" \
                " and \"azimuth\" in a future version.")            
            
        # initialise common structures
        self.mapped_quantities = mapped_quantities
        self.raw_mapping = {}
        self.mapping = {}
        
    def apply_mapping_functions(self, map_funcs={}, map_lims={}, param_lims={}):
        """ Taking input data and mapping to parameters.

        This function does the bulk of the work for `Source` classes,
        taking each input data variable and applying the mapping
        function (x' = x by default), descaling by the x' upper and
        lower limits and rescaling to the sonification parameter
        limits. These values are stored for non-evolving parameters,
        while for evolving properties they are converted to interpolation
        functions. 

        Args:
    	   map_funcs (:obj:`dict`, optional): dict with keys that must be
        	a subset self.mapped_quantities. Entries are then
        	function-like objects for converting input data
        	(e.g. taking log of a data set). If not provided,
        	each conversion function is assumed to be  f(x) = x.  
           map_lims (:obj:`dict`, optional): dict with keys that must be
        	a subset self.mapped_quantities. Entries are
        	tuples indicating the lower (index 0) and upper (index
        	1) limits on the converted input data
        	values. numerical values indicate absolute limits,
        	while strings are used to indicate percentiles
        	[e.g. ('10%','95%')]. converted data values are clipped
        	to these limits. If not provided, (0,1) is assumed.
           param_lims (:obj:`dict`, optional): dict with keys that
        	must be a subset self.mapped_quantities. Entries are
        	tuples indicating the lower (index 0) and upper (index
        	1) limits of the mapped sonification parameters. The
        	map_lims ranges are resaled to these ranges to give
        	the parameter values. If not provided, the default
        	param_lim_dict values are taken.

        Note:
           There is special behaviour for the `polar` and `azimuth`
           parameters, to ensure shortest angular distance when
           interpolating across the 0-2pi and 0-pi boundaries.
        
        """
        for key in self.mapped_quantities:
            rawvals = self.raw_mapping[key]

            # apply mapping functions if specified
            if key in map_funcs:
                mapvals = map_funcs[key](rawvals)
            else:
                mapvals = rawvals

            # set mapping limits if specified
            if key in map_lims:
                vallims = map_lims[key]
            else:
                vallims = (0,1)

            # set parameter limits if specified
            if key in param_lims:
                plims = param_lims[key]
            else:
                plims = param_lim_dict[key]
                
            lims = []
            # scale mapped values within limits if specified
            for l in vallims:
                if isinstance(l, str):
                    if '%' not in l:
                        warnings.warn("Specifying percentiles without appending a '%' character "
                                      "(e.g. XX%) currently works but is deprecated for more "
                                      "explicit syntax.", stacklevel=2)
                    else:
                        l = l.strip('%')
                    # string values notate percentile limits
                    pc = float(l)
                    buff = 1
                    sub = 0
                    if pc > 100:
                        # if percentile over 100 we add 
                        buff = pc/100.
                        pc = 100
                        sub = lims[0]
                    lim = sub + (np.percentile(np.hstack([mapvals]), pc) - sub)*buff
                    lims.append(lim)
                else:
                    # numerical values notate absolute limits
                    lims.append(l)
    
            # limit mapped values from 0 to 1 NOTE: do we want to mix and match const and evo?
            if hasattr(mapvals[0], "__iter__"):
                self.mapping[key] = []
                for i in range(self.n_sources):
                    scaledvals = rescale_values(mapvals[i], lims, plims)
                    self.mapping[key].append(scaledvals)
            else:
                scaledvals = rescale_values(np.array(mapvals), lims, plims)
                self.mapping[key] =  list(scaledvals)
            
        # finally, iterate through sources and interpolate evo functions 
        for key in self.mapping:
            if key == "time_evo":
                continue
            if key == "spectrum":
                # if hasattr(self.mapping[key][0][0], "__iter__"):
                # ^ in case we want to catch and pre process multi-spectra
                continue
            elif hasattr(self.mapping[key][0], "__iter__"):
                # print(key, self.mapping[key][0])
                for i in range(self.n_sources):
                    if key not in evolvable:
                        raise Exception(f"Mapping error: Parameter \"{key}\" cannot be evolved.")
                    x = self.mapping["time_evo"][i]
                    y = self.mapping[key][i]
                    if key == "phi" or key == "azimuth":
                        # special case: shortest angular distance
                        # between phi points is always assumed
                        ydiff = np.diff(y)
                        discont_bdx = abs(ydiff) > 0.5
                        for j in range(discont_bdx.sum()):
                            xpre = x[:-1][discont_bdx][j]
                            ysense = np.sign(ydiff[discont_bdx][j]) 
                            y[x > xpre] -= ysense
                    self.mapping[key][i] = interp1d(x,y, bounds_error=False,
                                                    fill_value=(y[0],y[-1]))
            
class Events(Source):
    """ Represent data as time-discrete events.

    Child class of `Source`, for `Event`-type sources. Each `Event` is
    discrete in `time` with single data values mapped to each
    sonification parameter.
    
    """
    def fromfile(self, datafile, coldict):
        """Take input data from ASCII file

        Args:
          datafile (:obj:`str`): path to input data file
          coldict (:obj:`dict`): keys are self.mapped_values, with
        	entries integer indexes for their corresponding column.
        """
        data = np.genfromtxt(datafile)
        for key in self.mapped_quantities:
            self.raw_mapping[key] = data[:,coldict[key]] 
        self.n_sources = data.shape[0]
        
    def fromdict(self, datadict):
        """Take input data from dictionary
	
        Args:
          datadict (:obj:`dict`): keys are self.mapped_values, with
        	entries corresponding to the input data. Multiple
        	sources are provided as :obj:`lists`, with data for
       		each source corresponding to the values. Single
        	sources can be represented as single values.
        """
        for key in self.mapped_quantities:
            if key in datadict:
                self.raw_mapping[key] = datadict[key]
            else:
                Exception(f"Mapped property {key} not in datadict.")
        self.n_sources = np.array(datadict[key]).shape[0]
 
class Objects(Source):
    """ Represent data as time-continuous objects.
    
    Child class of `Source`. In addition to supporting single values
    for each parameter (see `Events` class), objects also support
    time evolution for `evolvable` parameters, given a `time-evo`
    mapping.

    Todo:
    	* implement :obj:`fromfile` method
    """
    def fromdict(self, datadict):
        """ Take input data from dictionary
	
        Args:
          datadict (:obj:`dict`): keys are self.mapped_values, with
        	entries corresponding to the input data. Multiple
        	sources are provided as either :obj:`lists` or 2D
        	:obj:`numpy.array` objects, with each source
        	corresponding to the entries or columns respectively.
        	Single sources can be represented as single values or
        	1D :obj:`numpy.array` (for evolving parameters). 
        """
        for key in self.mapped_quantities:
            if key in datadict:
                d = datadict[key]
                if (type(d) is not list) and (np.array(d).ndim <= 1):
                    self.raw_mapping[key] = [d]
                else:
                    self.raw_mapping[key] = d
            else:
                Exception(f"Mapped property {key} not in datadict.")
        self.n_sources = np.array(self.raw_mapping[key]).shape[0]

class UnrecognisedProperty(Exception):
    "Error raised when trying to map unrecognised parameters"
    pass
